if __name__ is not None and "." in __name__:
    from .interval import Interval
else:
    from interval import Interval

if __name__ is not None and "." in __name__:
    from .pbox import Pbox
else:
    from pbox import Pbox

import scipy.stats as sps
import numpy as np

dists = {
    'alpha' : sps.alpha,
    'anglit' : sps.anglit,
    'arcsine' : sps.arcsine,
    'argus' : sps.argus,
    'beta' : sps.beta,
    'betaprime' : sps.betaprime,
    'bradford' : sps.bradford,
    'burr' : sps.burr,
    'burr12' : sps.burr12,
    'cauchy' : sps.cauchy,
    'chi' : sps.chi,
    'chi2' : sps.chi2,
    'cosine' : sps.cosine,
    'crystalball' : sps.crystalball,
    'dgamma' : sps.dgamma,
    'dweibull' : sps.dweibull,
    'erlang' : sps.erlang,
    'expon' : sps.expon,
    'exponnorm' : sps.exponnorm,
    'exponweib' : sps.exponweib,
    'exponpow' : sps.exponpow,
    'f' : sps.f,
    'fatiguelife' : sps.fatiguelife,
    'fisk' : sps.fisk,
    'foldcauchy' : sps.foldcauchy,
    'foldnorm' : sps.foldnorm,
    'frechet_r' : sps.frechet_r,
    'frechet_l' : sps.frechet_l,
    'genlogistic' : sps.genlogistic,
    'gennorm' : sps.gennorm,
    'genpareto' : sps.genpareto,
    'genexpon' : sps.genexpon,
    'genextreme' : sps.genextreme,
    'gausshyper' : sps.gausshyper,
    'gamma' : sps.gamma,
    'gengamma' : sps.gengamma,
    'genhalflogistic' : sps.genhalflogistic,
    'geninvgauss' : sps.geninvgauss,
    'gilbrat' : sps.gilbrat,
    'gompertz' : sps.gompertz,
    'gumbel_r' : sps.gumbel_r,
    'gumbel_l' : sps.gumbel_l,
    'halfcauchy' : sps.halfcauchy,
    'halflogistic' : sps.halflogistic,
    'halfnorm' : sps.halfnorm,
    'halfgennorm' : sps.halfgennorm,
    'hypsecant' : sps.hypsecant,
    'invgamma' : sps.invgamma,
    'invgauss' : sps.invgauss,
    'invweibull' : sps.invweibull,
    'johnsonsb' : sps.johnsonsb,
    'johnsonsu' : sps.johnsonsu,
    'kappa4' : sps.kappa4,
    'kappa3' : sps.kappa3,
    'ksone' : sps.ksone,
    'kstwobign' : sps.kstwobign,
    'laplace' : sps.laplace,
    'levy' : sps.levy,
    'levy_l' : sps.levy_l,
    'levy_stable' : sps.levy_stable,
    'logistic' : sps.logistic,
    'loggamma' : sps.loggamma,
    'loglaplace' : sps.loglaplace,
    'lognorm' : sps.lognorm,
    'loguniform' : sps.loguniform,
    'lomax' : sps.lomax,
    'maxwell' : sps.maxwell,
    'mielke' : sps.mielke,
    'moyal' : sps.moyal,
    'nakagami' : sps.nakagami,
    'ncx2' : sps.ncx2,
    'ncf' : sps.ncf,
    'nct' : sps.nct,
    'norm' : sps.norm,
    'norminvgauss' : sps.norminvgauss,
    'pareto' : sps.pareto,
    'pearson3' : sps.pearson3,
    'powerlaw' : sps.powerlaw,
    'powerlognorm' : sps.powerlognorm,
    'powernorm' : sps.powernorm,
    'rdist' : sps.rdist,
    'rayleigh' : sps.rayleigh,
    'rice' : sps.rice,
    'recipinvgauss' : sps.recipinvgauss,
    'semicircular' : sps.semicircular,
    'skewnorm' : sps.skewnorm,
    't' : sps.t,
    'trapz' : sps.trapz,
    'triang' : sps.triang,
    'truncexpon' : sps.truncexpon,
    'truncnorm' : sps.truncnorm,
    'tukeylambda' : sps.tukeylambda,
    'uniform' : sps.uniform,
    'vonmises' : sps.vonmises,
    'vonmises_line' : sps.vonmises_line,
    'wald' : sps.wald,
    'weibull_min' : sps.weibull_min,
    'weibull_max' : sps.weibull_max,
    'wrapcauchy' : sps.wrapcauchy,
    'bernoulli' : sps.bernoulli,
    'betabinom' : sps.betabinom,
    'binom' : sps.binom,
    'boltzmann' : sps.boltzmann,
    'dlaplace' : sps.dlaplace,
    'geom' : sps.geom,
    'hypergeom' : sps.hypergeom,
    'logser' : sps.logser,
    'nbinom' : sps.nbinom,
    'planck' : sps.planck,
    'poisson' : sps.poisson,
    'randint' : sps.randint,
    'skellam' : sps.skellam,
    'zipf' : sps.zipf,
    'yulesimon' : sps.yulesimon
}

def __get_bounds(function_name = None,steps = 200,*args,):

    x = np.linspace(0.0001,0.9999,steps)

    arg_list = [list(args)]
    complete = False
    # while not complete:
    for i in arg_list:
        for j in range(len(i)):
            if i[j].__class__.__name__ == 'Interval':

                k1 = [i[j].left() if l == i[j] else l for l in i]
                k2 = [i[j].right() if l == i[j] else l for l in i]

                arg_list.append(k1)
                arg_list.append(k2)

    new_args = []
    for k in arg_list:
        to_delete = False

        for l in k:

            if l.__class__.__name__ == 'Interval':
                to_delete = True

        if not to_delete and k not in new_args:
            new_args.append(k)

    bounds = []

    mean_hi = -np.inf
    mean_lo = np.inf
    var_lo = np.inf
    var_hi = 0

    for a in new_args:

        bounds.append(dists[function_name].ppf(x,*a))
        bmean, bvar = dists[function_name].stats(*a, moments = 'mv')

        if bmean < mean_lo:
            mean_lo = bmean
        if bmean > mean_hi:
            mean_hi = bmean
        if bvar > var_hi:
            var_hi = bvar
        if bvar < var_lo:
            var_lo = bvar


    Left = [min([b[i] for b in bounds]) for i in range(steps)]
    Right = [max([b[i] for b in bounds]) for i in range(steps)]

    var  = Interval(np.float64(var_lo),np.float64(var_hi))
    mean = Interval(np.float64(mean_lo),np.float64(mean_hi))

    Left = np.array(Left)
    Right = np.array(Right)

    return Left, Right, mean, var


def lognormal(mean, var, steps = 200):

    x = np.linspace(0,1,steps)

    if mean.__class__.__name__ != 'Interval':
        mean = Interval(mean,mean)
    if var.__class__.__name__ != 'Interval':
        var = Interval(var,var)

    bound0 = sps.lognorm.ppf(x, mean.left(), var.left())
    bound1 = sps.lognorm.ppf(x, mean.right(), var.left())
    bound2 = sps.lognorm.ppf(x, mean.left(), var.right())
    bound3 = sps.lognorm.ppf(x, mean.right(), var.right())

    Left = [min(bound0[i],bound1[i],bound2[i],bound3[i]) for i in range(steps)]
    Right = [max(bound0[i],bound1[i],bound2[i],bound3[i]) for i in range(steps)]

    Left = np.array(Left)
    Right = np.array(Right)
    return Pbox(
        Left,
        Right,
        steps = steps,
        shape='normal',
        mean_left=mean.left(),
        mean_right=mean.right(),
        var_left=var.left(),
        var_right=var.right())
lognorm = lognormal

def alpha(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('alpha',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'alpha',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def anglit(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('anglit',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'anglit',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def arcsine(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('arcsine',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'arcsine',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def argus(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('argus',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'argus',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def beta(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('beta',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'beta',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def betaprime(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('betaprime',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'betaprime',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def bradford(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('bradford',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'bradford',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def burr(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('burr',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'burr',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def burr12(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('burr12',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'burr12',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def cauchy(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('cauchy',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'cauchy',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def chi(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('chi',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'chi',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def chi2(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('chi2',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'chi2',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def cosine(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('cosine',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'cosine',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def crystalball(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('crystalball',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'crystalball',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def dgamma(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('dgamma',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'dgamma',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def dweibull(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('dweibull',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'dweibull',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def erlang(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('erlang',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'erlang',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def expon(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('expon',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'expon',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def exponnorm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('exponnorm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'exponnorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def exponweib(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('exponweib',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'exponweib',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def exponpow(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('exponpow',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'exponpow',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def f(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('f',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'f',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def fatiguelife(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('fatiguelife',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'fatiguelife',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def fisk(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('fisk',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'fisk',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def foldcauchy(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('foldcauchy',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'foldcauchy',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def foldnorm(mu,s, steps = 200):

    x = np.linspace(0.0001,0.9999,steps)
    if mu.__class__.__name__ != 'Interval':
        mu = Interval(mu)
    if s.__class__.__name__ != 'Interval':
        s = Interval(s)

    new_args = [
        [mu.lo()/s.lo(),0,s.lo()],
        [mu.hi()/s.lo(),0,s.lo()],
        [mu.lo()/s.hi(),0,s.hi()],
        [mu.hi()/s.hi(),0,s.hi()]
    ]


    bounds = []

    mean_hi = -np.inf
    mean_lo = np.inf
    var_lo = np.inf
    var_hi = 0

    for a in new_args:

        bounds.append(sps.foldnorm.ppf(x,*a))
        bmean, bvar = sps.foldnorm.stats(*a, moments = 'mv')

        if bmean < mean_lo:
            mean_lo = bmean
        if bmean > mean_hi:
            mean_hi = bmean
        if bvar > var_hi:
            var_hi = bvar
        if bvar < var_lo:
            var_lo = bvar


    Left = [min([b[i] for b in bounds]) for i in range(steps)]
    Right = [max([b[i] for b in bounds]) for i in range(steps)]

    var  = Interval(np.float64(var_lo),np.float64(var_hi))
    mean = Interval(np.float64(mean_lo),np.float64(mean_hi))

    Left = np.array(Left)
    Right = np.array(Right)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'foldnorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def frechet_r(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('frechet_r',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'frechet_r',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def frechet_l(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('frechet_l',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'frechet_l',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def genlogistic(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('genlogistic',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'genlogistic',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def gennorm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('gennorm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'gennorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def genpareto(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('genpareto',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'genpareto',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def genexpon(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('genexpon',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'genexpon',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def genextreme(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('genextreme',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'genextreme',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def gausshyper(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('gausshyper',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'gausshyper',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def gamma(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('gamma',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'gamma',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def gengamma(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('gengamma',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'gengamma',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def genhalflogistic(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('genhalflogistic',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'genhalflogistic',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def geninvgauss(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('geninvgauss',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'geninvgauss',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def gilbrat(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('gilbrat',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'gilbrat',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def gompertz(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('gompertz',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'gompertz',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def gumbel_r(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('gumbel_r',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'gumbel_r',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def gumbel_l(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('gumbel_l',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'gumbel_l',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def halfcauchy(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('halfcauchy',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'halfcauchy',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def halflogistic(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('halflogistic',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'halflogistic',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def halfnorm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('halfnorm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'halfnorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def halfgennorm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('halfgennorm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'halfgennorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def hypsecant(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('hypsecant',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'hypsecant',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def invgamma(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('invgamma',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'invgamma',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def invgauss(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('invgauss',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'invgauss',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def invweibull(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('invweibull',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'invweibull',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def johnsonsb(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('johnsonsb',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'johnsonsb',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def johnsonsu(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('johnsonsu',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'johnsonsu',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def kappa4(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('kappa4',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'kappa4',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def kappa3(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('kappa3',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'kappa3',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def ksone(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('ksone',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'ksone',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def kstwobign(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('kstwobign',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'kstwobign',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def laplace(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('laplace',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'laplace',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def levy(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('levy',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'levy',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def levy_l(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('levy_l',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'levy_l',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def levy_stable(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('levy_stable',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'levy_stable',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def logistic(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('logistic',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'logistic',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def loggamma(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('loggamma',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'loggamma',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def loglaplace(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('loglaplace',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'loglaplace',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def lognorm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('lognorm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'lognorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def loguniform(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('loguniform',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'loguniform',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def lomax(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('lomax',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'lomax',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def maxwell(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('maxwell',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'maxwell',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def mielke(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('mielke',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'mielke',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def moyal(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('moyal',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'moyal',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def nakagami(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('nakagami',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'nakagami',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def ncx2(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('ncx2',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'ncx2',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def ncf(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('ncf',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'ncf',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def nct(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('nct',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'nct',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def norm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('norm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'norm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def norminvgauss(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('norminvgauss',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'norminvgauss',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def pareto(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('pareto',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'pareto',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def pearson3(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('pearson3',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'pearson3',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def powerlaw(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('powerlaw',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'powerlaw',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def powerlognorm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('powerlognorm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'powerlognorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def powernorm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('powernorm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'powernorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def rdist(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('rdist',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'rdist',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def rayleigh(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('rayleigh',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'rayleigh',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def rice(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('rice',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'rice',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def recipinvgauss(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('recipinvgauss',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'recipinvgauss',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def semicircular(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('semicircular',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'semicircular',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def skewnorm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('skewnorm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'skewnorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def t(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('t',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 't',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def trapz(a,b,c,d , steps = 200):
    if a.__class__.__name__ != 'Interval':
        a = Interval(a)
    if b.__class__.__name__ != 'Interval':
        b = Interval(b)
    if c.__class__.__name__ != 'Interval':
        c = Interval(c)
    if d.__class__.__name__ != 'Interval':
        d = Interval(d)

    x = np.linspace(0.0001,0.9999,steps)
    left = sps.trapz.ppf(x,b.lo()/d.lo(),c.lo()/d.lo(),a.lo(),d.lo()-a.lo())
    right = sps.trapz.ppf(x,b.hi()/d.hi(),c.hi()/d.hi(),a.hi(),d.hi()-a.hi())

    return Pbox(
          left,
          right,
          steps      = steps,
          shape      = 'trapz'
          )

def triang(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('triang',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'triang',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def truncexpon(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('truncexpon',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'truncexpon',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def truncnorm(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('truncnorm',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'truncnorm',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def tukeylambda(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('tukeylambda',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'tukeylambda',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )


def uniform(a, b, steps = 200):

    x = np.linspace(0,1,steps)

    if a.__class__.__name__ != 'Interval':
        a = Interval(a,a)
    if b.__class__.__name__ != 'Interval':
        b = Interval(b,b)

    Left = np.linspace(a.left(),b.left())
    Right = np.linspace(a.right(),b.right())

    mean = 0.5 * (a+b)
    var = ((b-a)**2 )/12

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'uniform',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def vonmises(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('vonmises',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'vonmises',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def vonmises_line(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('vonmises_line',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'vonmises_line',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def wald(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('wald',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'wald',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def weibull_min(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('weibull_min',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'weibull_min',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def weibull_max(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('weibull_max',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'weibull_max',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def wrapcauchy(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('wrapcauchy',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'wrapcauchy',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def bernoulli(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('bernoulli',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'bernoulli',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def betabinom(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('betabinom',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'betabinom',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def binom(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('binom',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'binom',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def boltzmann(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('boltzmann',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'boltzmann',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def dlaplace(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('dlaplace',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'dlaplace',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def geom(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('geom',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'geom',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def hypergeom(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('hypergeom',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'hypergeom',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def logser(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('logser',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'logser',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def nbinom(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('nbinom',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'nbinom',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def planck(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('planck',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'planck',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def poisson(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('poisson',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'poisson',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def randint(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('randint',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'randint',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def skellam(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('skellam',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'skellam',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def zipf(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('zipf',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'zipf',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )

def yulesimon(*args, steps = 200):
    args = list(args)
    for i in range(0,len(args)):
        if args[i].__class__.__name__ != 'Interval':
            args[i] = Interval(args[i])

    Left, Right, mean, var = __get_bounds('yulesimon',steps,*args)

    return Pbox(
          Left,
          Right,
          steps      = steps,
          shape      = 'yulesimon',
          mean_left  = mean.left(),
          mean_right = mean.right(),
          var_left   = var.left(),
          var_right  = var.right()
          )


### Other distributions
def KM(k,m,steps = 200):
    return beta(Interval(k,k+1),Interval(m,m+1),steps = steps)

def KN(k,n,steps = 200):
    return KM(k,n-k,steps=steps)


### Alternate names
normal = norm
N = normal
unif = uniform
U = uniform



import numpy as np
from matplotlib import pyplot as plt

from .interval import Interval

class Pbox(object):

    def __init__(self, left=None, right=None, steps=200, shape=None, mean_left=None, mean_right=None, var_left=None, var_right=None, interpolation='linear'):

        if (left is not None) and (right is None):
            right = left

        if left is None and right is None:
            left = -np.inf
            right = np.inf

        if isinstance(left, Interval):
            left = np.array([left.left()])

        if isinstance(right, Interval):
            right = np.array([right.right()])

        if len(left) != steps:
            left = interpolate(left, interpolation=interpolation, left=False, steps=steps)

        if len(right) != steps:
            right = interpolate(right, interpolation=interpolation, left=True, steps=steps)

        self.left = left
        self.right = right

        self.steps = steps
        self.n = self.steps
        self.shape = shape
        self.mean_left = -np.inf
        self.mean_right = np.inf
        self.var_left = 0
        self.var_right = np.inf

        self._computemoments()
        if shape is not None: self.shape = shape
        if mean_left is not None: self.mean_left = np.max([mean_left, self.mean_left])
        if mean_right is not None: self.mean_right = np.min([mean_right, self.mean_right])
        if var_left is not None: self.var_left = np.max([var_left, self.var_left])
        if var_right is not None: self.var_right = np.min([var_right, self.var_right])
        self._checkmoments()

    def __repr__(self):
        if self.mean_left == self.mean_right:
            mean_text = f'{round(self.mean_left, 4)}'
        else:
            mean_text = f'[{round(self.mean_left, 4)}, {round(self.mean_right, 4)}]'

        if self.var_left == self.var_right:
            var_text = f'{round(self.var_left, 4)}'
        else:
            var_text = f'[{round(self.var_left, 4)}, {round(self.var_right, 4)}]'

        range_text = f'[{round(np.min([self.left, self.right]), 4), round(np.max([self.left, self.right]), 4)}'

        if self.shape is None:
            shape_text = ' '
        else:
            shape_text = f' {self.shape}' # space to start; see below lacking space

        return f'Pbox: ~{shape_text}(range={range_text}, mean={mean_text}, var={var_text})'

    def __iter__(self):
        for val in np.array([self.left,self.right]).flatten():
            yield val

    def __neg__(self):
        if self.shape in ['uniform','normal','cauchy','triangular','skew-normal']:
            s = self.shape
        else:
            s = ''

        return Pbox(
            left = -np.flip(self.right),
            right = -np.flip(self.left),
            shape = s,
            mean_left = -self.mean_right,
            mean_right = -self.mean_left,
            var_left = self.var_left,
            var_right = self.var_right
        )

    def __add__(self, other):

        return self.add(other, method = 'f')

    def __radd__(self,other):
        return self.add(other, method = 'f')

    def __sub__(self,other):

        return self.sub(other, method = 'f')

    def __rsub__(self,other):
        self = - self
        return self.add(other, method = 'f')

    def __mul__(self,other):
        return self.mul(other, method = 'f')

    def __rmul__(self,other):
        return self.mul(other, method = 'f')

    def __truediv__(self, other):

        return self.div(other, method = 'f')

    def __rtruediv__(self,other):

        try:
            return other * self.recip()
        except:
            return NotImplemented

    ### Local functions ###
    def _computemoments(self):    # should we compute mean if it is a Cauchy, var if it's a t distribution?
        self.mean_left = np.max([self.mean_left, np.mean(self.left)])
        self.mean_right = np.min([self.mean_right, np.mean(self.right)])

        if not (np.any(self.left <= -np.inf) or np.any(np.inf <= self.right)):
            V, JJ = 0, 0
            j = np.array(range(self.n))

            for J in np.array(range(self.n)) - 1:
                ud = [*self.left[j < J], *self.right[J <= j]]
                v = sideVariance(ud)

                if V < v:
                    JJ = J
                    V = v

            self.var_right = V

    def _checkmoments(self):

        a = Interval(self.mean_left, self.mean_right) #mean(x)
        b = dwMean(self)

        self.mean_left = np.max([left(a), left(b)])
        self.mean_right = np.min([right(a), right(b)])

        if self.mean_right < self.mean_left:
            # use the observed mean
            self.mean_left = left(b)
            self.mean_right = right(b)

        a = Interval(self.var_left, self.var_right) #var(x)
        b = dwVariance(self)

        self.var_left = np.max([left(a), left(b)])
        self.var_right = np.min([right(a),right(b)])

        if self.var_right < self.var_left:
            # use the observed variance
            self.var_left = left(b)
            self.var_right = right(b)

    ### Public funtions ###
    def add(self, other, method = 'f'):

        if method not in ['f','p','o','i']:
            raise ArithmeticError("Calculation method unkown")

        if other.__class__.__name__ == 'Interval':
            other = Pbox(other, steps = self.steps)

        if other.__class__.__name__ == 'Pbox':

            if self.steps != other.steps:
                raise ArithmeticError("Both Pboxes must have the same number of steps")

            if method == 'f':

                nleft  = np.empty(self.steps)
                nright = np.empty(self.steps)

                for i in range(0,self.steps):
                    j = np.array(range(i, self.steps))
                    k = np.array(range(self.steps - 1, i-1, -1))

                    nleft[i] = np.min(self.right[j] + other.right[k])

                    jj = np.array(range(0, i + 1))
                    kk = np.array(range(i, -1 , -1))

                    nright[i] = np.max(self.left[jj] + other.left[kk])

            elif method == 'p':

                nleft  = self.left + other.left
                nright = self.right + other.right

            elif method == 'o':

                nleft  = self.left + np.flip(other.left)
                nright = self.right + np.flip(other.right)

            elif method == 'i':

                nleft  = []
                nright = []
                for i in self.left:
                    for j in other.left:
                        nleft.append(i+j)
                for ii in self.right:
                    for jj in other.right:
                        nright.append(ii+jj)

            nleft.sort()
            nright.sort()

            return Pbox(
                left    = nleft,
                right   = nright,
                steps   = self.steps
            )

        else:
            try:
                # Try adding constant
                if self.shape in ['uniform','normal','cauchy','triangular','skew-normal']:
                    s = self.shape
                else:
                    s = ''

                return Pbox(
                    left       = self.left + other,
                    right      = self.right + other,
                    shape      = s,
                    mean_left  = self.mean_left + other,
                    mean_right = self.mean_right + other,
                    var_left   = self.var_left,
                    var_right  = self.var_right,
                    steps      = self.steps
                )

            except:
                return NotImplemented

    def sub(self, other, method = 'f'):

        if method == 'o':
            method = 'p'
        elif method == 'p':
            method = 'o'

        return self.add(-other, method)

    def mul(self, other, method = 'f'):

        if method not in ['f','p','o','i']:
            raise ArithmeticError("Calculation method unkown")

        if other.__class__.__name__ == 'Interval':
            other = Pbox(other, steps = self.steps)

        if other.__class__.__name__ == 'Pbox':

            if self.steps != other.steps:
                raise ArithmeticError("Both Pboxes must have the same number of steps")

            if method == 'f':

                nleft  = np.empty(self.steps)
                nright = np.empty(self.steps)

                for i in range(0,self.steps):
                    j = np.array(range(i, self.steps))
                    k = np.array(range(self.steps - 1, i-1, -1))

                    nleft[i] = np.min(self.right[j] * other.right[k])

                    jj = np.array(range(0, i + 1))
                    kk = np.array(range(i, -1 , -1))

                    nright[i] = np.max(self.left[jj] * other.left[kk])

            elif method == 'p':

                nleft  = self.left * other.left
                nright = self.right * other.right

            elif method == 'o':

                nleft  = self.left * np.flip(other.left)
                nright = self.right * np.flip(other.right)

            elif method == 'i':

                nleft  = []
                nright = []
                for i in self.left:
                    for j in other.left:
                        nleft.append(i*j)
                for ii in self.right:
                    for jj in other.right:
                        nright.append(ii*jj)

            nleft.sort()
            nright.sort()

            return Pbox(
                left    = nleft,
                right   = nright,
                steps   = self.steps
            )

        else:
            try:
                # Try adding constant
                if self.shape in ['uniform','normal','cauchy','triangular','skew-normal']:
                    s = self.shape
                else:
                    s = ''

                return Pbox(
                    left       = self.left * other,
                    right      = self.right * other,
                    shape      = s,
                    mean_left  = self.mean_left * other,
                    mean_right = self.mean_right * other,
                    var_left   = self.var_left,
                    var_right  = self.var_right,
                    steps      = self.steps
                )

            except:
                return NotImplemented

    def div(self, other, method = 'f'):

        if method == 'o':
            method = 'p'
        elif method == 'p':
            method = 'o'

        return self.mul(1/other, method)

    def recip(self):
        return Pbox(
            left  = 1 / np.flip(self.right),
            right = 1 / np.flip(self.left),
            steps = self.steps
        )

    def show(self,now = True,**kwargs):
        # If you want to know why numpy is the WORST thing about numpy
        # see this code snippet
        left, right = self.get_x()
        y  = self.get_y()


        plt.plot(left,y,**kwargs)
        plt.plot(right,y,**kwargs)
        if now:
            plt.show()
        else:
            return plt

    def get_interval(self, *args):

        if len(args) == 1:

            if args[0] == 1:
                # asking for whole pbox bounds
                return Interval(min(self.left),max(self.right))

            p1 = (1-args[0])/2
            p2 = 1-p1

        elif len(args) == 2:

            p1 = args[0]
            p2 = args[1]

        else:
            raise Exception('Too many inputs')

        y  = np.append(np.insert(np.linspace(0,1,self.steps),0,0),1)

        y1 = 0
        while y[y1] < p1:
            y1 += 1

        y2 = len(y)-1
        while y[y2] > p2:
            y2 -= 1

        x1 = self.left[y1]
        x2 = self.right[y2]
        return Interval(x1,x2)

    def get_probability(self, val):
        p  = np.append(np.insert(np.linspace(0,1,self.steps),0,0),1)

        i = 0
        while i < self.steps and self.left[i] < val:
            i += 1


        ub = p[i]

        j = 0

        while j < self.steps and self.right[j] < val:
            j += 1


        lb = p[j]

        return Interval(lb,ub)

    def support(self):
        return np.linspace(0,1,self.steps)

    def get_x(self):
        # returns the x values for plotting
        left = np.append(np.insert(self.left,0,min(self.left)),max(self.right))
        right = np.append(np.insert(self.right,0,min(self.left)),max(self.right))
        return left, right

    def get_y(self):
        # returns y values for plotting
        return np.append(np.insert(np.linspace(0,1,self.steps),0,0),1)
# Public functions

# Funcitons
def env_int(*args):
    left = min([min(i) if is_iterable(i) else i for i in args])
    right = max([max(i) if is_iterable(i) else i for i in args])
    return Interval(left, right)

def left(imp):
    if isinstance(imp, Interval) or isinstance(imp, pbox.Pbox):
        return imp.left()
    elif is_iterable(imp):
        return min(imp)
    else:
        return imp

def right(imp):
    if isinstance(imp, Interval) or isinstance(imp, pbox.Pbox):
        return imp.right()
    elif is_iterable(imp):
        return max(imp)
    else:
        return imp

def left_list(implist, verbose=False):
    if not is_iterable(implist):
        return np.array(implist)

    return np.array([left(imp) for imp in implist])

def right_list(implist, verbose=False):
    if not is_iterable(implist):
        return np.array(implist)

    return np.array([right(imp) for imp in implist])

def qleftquantiles(pp, x, p): # if first p is not zero, the left tail will be -Inf
    return [max(left_list(x)[right_list(p) <= P]) for P in pp]

def qrightquantiles(pp, x, p):  # if last p is not one, the right tail will be Inf
    return [min(right_list(x)[P <= left_list(p)]) for P in pp]

def quantiles(x, p, steps=200):
    left = qleftquantiles(ii(steps=steps), x, p)
    right = qrightquantiles(jj(steps=steps), x, p)
    return pbox.Pbox(left=left, right=right)  # quantiles are in x and the associated cumulative probabilities are in p

def interp_step(u, steps=200):
    u = np.sort(u)

    seq = np.linspace(start=0, stop=len(u) - 0.00001, num=steps, endpoint=True)
    seq = np.array([trunc(seq_val) for seq_val in seq])
    return u[seq]

def interp_cubicspline(vals, steps=200):
    vals = np.sort(vals) # sort
    vals_steps = np.array(range(len(vals))) + 1
    vals_steps = vals_steps / len(vals_steps)

    steps = np.array(range(steps)) + 1
    steps = steps / len(steps)

    interped = interp.CubicSpline(vals_steps, vals)
    return interped(steps)

def interp_left(u, steps=200):
    p = np.array(range(len(u))) / (len(u) - 1)
    pp, x = ii(steps=steps), u
    return qleftquantiles(pp, x, p)

def interp_right(d, steps=200):
    p = np.array(range(len(d))) / (len(d) - 1)
    pp, x = jj(steps=steps), d
    return qrightquantiles(pp, x, p)

def interp_outer(x, left, steps=200):
    if (left) :
        return interp_left(x, steps=steps)
    else:
        return interp_right(x, steps=steps)

def interp_linear(V, steps=200):
    m = len(V) - 1

    if m == 0: return np.repeat(V, steps)
    if steps == 1: return np.array([min(V), max(V)])

    d = 1 / m
    n = round(d * steps * 20)

    if n == 0:
        c = V
    else:
        c = []
        for i in range(m):
            v = V[i]
            w = V[i + 1]
            c.extend(np.linspace(start=v, stop=w, num=n))

    u = [c[round((len(c) - 1) * (k + 0) / (steps - 1))] for k in range(steps)]

    return np.array(u)

def interpolate(u, interpolation='linear', left=True, steps=200):
    if interpolation == 'outer':
        return interp_outer(u, left, steps=steps)
    elif interpolation == 'spline':
        return interp_cubicspline(u, steps=steps)
    elif interpolation == 'step':
        return interp_step(u, steps=steps)
    else:
        return interp_linear(u, steps=steps)

def sideVariance(w, mu=None):
    if not isinstance(w, np.ndarray): w = np.array(w)
    if mu is None: mu = np.mean(w)
    return max(0, np.mean((w - mu) ** 2))

def dwMean(pbox):
    return Interval(np.mean(pbox.right), np.mean(pbox.left))

def dwVariance(pbox):
    if np.any(np.isinf(pbox.left)) or np.any(np.isinf(pbox.right)):
        return Interval(0, np.inf)

    if np.all(pbox.right[0] == pbox.right) and np.all(pbox.left[0] == pbox.left):
        return Interval(0, (pbox.right[0] - pbox.left[0]) ** (2 / 4))

    vr = sideVariance(pbox.left, np.mean(pbox.left))
    w = np.copy(pbox.left)
    n = len(pbox.left)

    for i in reversed(range(n)):
        w[i] = pbox.right[i]
        v = sideVariance(w, np.mean(w))

        if np.isnan(vr) or np.isnan(v):
            vr = np.inf
        elif vr < v:
            vr = v

    if pbox.left[n - 1] <= pbox.right[0]:
        vl = 0.0
    else:
        x = pbox.right
        vl = sideVariance(w, np.mean(w))

        for i in reversed(range(n)):
            w[i] = pbox.left[i]
            here = w[i]

            if 1 < i:
                for j in reversed(range(i-1)):
                    if w[i] < w[j]:
                        w[j] = here

            v = sideVariance(w, np.mean(w))

            if np.isnan(vl) or np.isnan(v):
                vl = 0
            elif v < vl:
                vl = v

    return Interval(vl, vr)

def straddles(x):
    return (left(x) <= 0) and (0 <= right(x)) # includes zero

def straddlingzero(x):
    return (left(x) < 0) and (0 < right(x)) # neglects zero as an endpoint



# -*- coding: utf-8 -*-
"""
Created on Thu Oct 17 12:57:55 2019

@author: sggdale (with 'inspiration' from Marco)

nick: Sorry
"""

#testcommit

import numpy as np
import random as r


class Interval():

    def __repr__(self): # return
        return "[%g, %g]"%(self.Left,self.Right)

    def __str__(self): # print
        return "[%g, %g]"%(self.Left,self.Right)

    def __init__(self,Left = None, Right = None):

        # kill complex nums
        assert not isinstance(Left, np.complex) or not isinstance(Right, np.complex), "Inputs must be real numbers"

        # assume vaccous if no inputs
        if Left is None and Right is None:
            Right = np.inf
            Left = np.inf

        # If only one input assume zero width
        elif Left is None and Right is not None:
            Left = Right
        elif Left is not None and Right is None:
            Right = Left

        # if iterable, find endpoints
        if hasattr(Left, '__iter__') and hasattr(Right, '__iter__'):

            LL = min(Left)
            UL = min(Right)
            LU = max(Left)
            UU = max(Right)

            Left = min(LL,LU)
            Right = max(LU,UU)

        elif hasattr(Left, '__iter__'):

            LL = min(Left)
            LU = max(Left)

            Left = min(LL,LU)


        elif hasattr(Right, '__iter__'):

            UL = min(Right)
            UU = max(Right)

            Right = max(LU,UU)


        if Left > Right:
            LowerUpper = [Left, Right]
            Left = min(LowerUpper)
            Right = max(LowerUpper)

        self.Left = Left
        self.Right = Right

    def __iter__(self):
        for bound in [self.Left, self.Right]:
            yield bound

    def __len__(self):
        return 2

    def __add__(self,other):

        if other.__class__.__name__ == 'Interval':
            lo = self.Left + other.Left
            hi = self.Right + other.Right
        elif other.__class__.__name__ == 'Pbox':
            # Perform Pbox addition assuming independance
            return other.add(self, method = 'i')
        else:
            try:
                lo = self.Left + other
                hi  = self.Right + other
            except:
                return NotImplemented

        return Interval(lo,hi)

    def __radd__(self,left):
        return self.__add__(left)

    def __sub__(self, other):

        if other.__class__.__name__ == "Interval":

            lo = self.Left - other.Right
            hi = self.Right - other.Left
        elif other.__class__.__name__ == "Pbox":
            # Perform Pbox subtractnion assuming independance
            return other.rsub(self)
        else:
            try:
                lo = self.Left - other
                hi  = self.Right - other
            except:
                return NotImplemented

        return Interval(lo,hi)

    def __rsub__(self, other):
        if other.__class__.__name__ == "Interval":
            # should be overkill
            lo = other.Right - self.Left
            hi = other.Right - self.Right

        elif other.__class__.__name__ == "Pbox":
            # shoud have be caught by Pbox.__sub__()
            return other.__sub__(self)
        else:
            try:
                lo = other - self.Right
                hi = other - self.Left

            except:
                return NotImplemented

        return Interval(lo,hi)

    def __mul__(self,other):
        if other.__class__.__name__ == "Interval":

            b1 = self.lo() * other.lo()
            b2 = self.lo() * other.hi()
            b3 = self.hi() * other.lo()
            b4 = self.hi() * other.hi()

            lo = min(b1,b2,b3,b4)
            hi = max(b1,b2,b3,b4)

        elif other.__class__.__name__ == "Pbox":

            return other.mul(self)

        else:

            try:

                lo = self.lo() * other
                hi = self.hi() * other

            except:
                
                return NotImplemented

        return Interval(lo,hi)

    def __rmul__(self,other):
        return self * other

    def __truediv__(self,other):

        if other.__class__.__name__ == "Interval":

            if other.straddles_zero():
                # Cant divide by zero
                raise ZeroDivisionError()

            b1 = self.lo() / other.lo()
            b2 = self.lo() / other.hi()
            b3 = self.hi() / other.lo()
            b4 = self.hi() / other.hi()

            lo = min(b1,b2,b3,b4)
            hi = max(b1,b2,b3,b4)

        else:
            try:
                lo = self.lo()/other
                hi = self.hi()/other
            except:

                return NotImplemented

        return Interval(lo,hi)


    def __rtruediv__(self,other):

        try:
            return other * self.recip()
        except:
            return NotImplemented


    def __pow__(self,other):
        if other.__class__.__name__ == "Interval":
            pow1 = self.Left ** other.Left
            pow2 = self.Left ** other.Right
            pow3 = self.Right ** other.Left
            pow4 = self.Right ** other.Right
            powUp = max(pow1,pow2,pow3,pow4)
            powLow = min(pow1,pow2,pow3,pow4)
        elif other.__class__.__name__ in ("int", "float"):
            pow1 = self.Left ** other
            pow2 = self.Right ** other
            powUp = max(pow1,pow2)
            powLow = min(pow1,pow2)
            if (self.Right >= 0) and (self.Left <= 0) and (other % 2 == 0):
                powLow = 0
        return Interval(powLow,powUp)

    def __rpow__(self,left):
        if left.__class__.__name__ == "Interval":
            pow1 = left.Left ** self.Left
            pow2 = left.Left ** self.Right
            pow3 = left.Right ** self.Left
            pow4 = left.Right ** self.Right
            powUp = max(pow1,pow2,pow3,pow4)
            powLow = min(pow1,pow2,pow3,pow4)

        elif left.__class__.__name__ in ("int", "float"):
            pow1 = left ** self.Left
            pow2 = left ** self.Right
            powUp = max(pow1,pow2)
            powLow = min(pow1,pow2)

        return Interval(powLow,powUp)


    def left(self):
        return self.Left

    def right(self):
        return self.Right

    lo = left
    hi = right

    def mean(*args):
        LSum = 0
        USum = 0
        DataLen = len(args)
        for x in args:
            if x.__class__.__name__ in ("int","float"):
                x = Interval(x)
            if x.__class__.__name__ in ("list","tuple"):
                DataLen = DataLen + (len(x) - 1)
                for y in x:
                    if y.__class__.__name__ in ("int","float"):
                        y = Interval(y)
                    LSum = LSum + y.Left
                    USum = USum + y.Right
            if x.__class__.__name__ == "Interval":
                LSum = LSum + x.Left
                USum = USum + x.Right
            LMean = LSum / DataLen
            UMean = USum / DataLen
        return Interval(LMean, UMean)

    def median(*args):
        LBounds = []
        LSorted = []
        UBounds = []
        USorted = []

        for x in [*args]:
            if x.__class__.__name__ in ("int","float"):
                x = Interval(x)
                LBounds.append(x.Left)
                UBounds.append(x.Right)
            if x.__class__.__name__ in ("list","tuple"):
                for y in x:
                    if y.__class__.__name__ in ("int","float"):
                        y = Interval(y)
                    LBounds.append(y.Left)
                    UBounds.append(y.Right)
            if x.__class__.__name__ == "Interval":
                LBounds.append(x.Left)
                UBounds.append(x.Right)
        while (len(LBounds) > 0):
            MinL = min(LBounds)
            LSorted.append(MinL)
            LBounds.remove(MinL)
        while (len(UBounds) > 0):
            MinU = min(UBounds)
            USorted.append(MinU)
            UBounds.remove(MinU)

        if (len(LSorted) % 2) != 0:
            LMedian = LSorted[len(LSorted)//2]
            UMedian = USorted[len(USorted)//2]
        else:
            LMedian = (LSorted[len(LSorted)//2] + LSorted[(len(LSorted)//2)-1])/2
            UMedian = (USorted[len(USorted)//2] + USorted[(len(USorted)//2)-1])/2
        return Interval(LMedian,UMedian)

    def variance(*args):
        dataMean = Interval.mean(*args)
        LBounds = []
        UBounds = []
        LDev = []
        UDev = []
        DataLen = len(args)
        for x in [*args]:
            if x.__class__.__name__ in ("int","float"):
                x = Interval(x)
                LBounds.append(x.Left)
                UBounds.append(x.Right)
            if x.__class__.__name__ in ("list","tuple"):
                DataLen = DataLen + (len(x) - 1)
                for y in x:
                    if y.__class__.__name__ in ("int","float"):
                        y = Interval(y)
                    LBounds.append(y.Left)
                    UBounds.append(y.Right)

        for y in LBounds:
            LDev.append(abs(y - dataMean.Left)**2)
        for z in UBounds:
            UDev.append(abs(z - dataMean.Right)**2)

        LSDev = (sum(LDev))/DataLen
        USDev = (sum(UDev))/DataLen
        return Interval(LSDev, USDev)

    def mode(*args):
        NotImplemented

    def straddles(self,N):
        if self.Left <= N and self.Right >= N:
            return True
        else:
            return False

    def straddles_zero(self):
        self.straddles(0)


    def recip(self):

        if self.straddles_zero():
            # Cant divide by zero
            raise ZeroDivisionError()

        elif 1/self.hi() < 1/self.lo():
            return Interval(1/self.hi(), 1/self.lo())
        else:
            return Interval(1/self.lo(), 1/self.hi())


# a = Interval(1,2)
# b = Interval(3,4)
# c = Interval(-2,5)
# d = Interval(-7,-4)

##.sort() function to sort numbers for median
##list1.count(x) function to help with mode


I = Interval



from distutils.core import setup

setup(
    name='pba',
    version='0.5.7',
    packages=['pba',],
    license='MIT License',
    long_description=open('README.rst').read(),
    long_description_content_type="text/markdown",
    install_requires=[
        'numpy>=1.17.3',
        'scipy>=1.4.0',
        'matplotlib>=3.0.0'],
    python_requires='>=3',
    url='https://github.com/ngg1995/pba-for-python',
    author='Nick Gray',
    author_email = 'nickgray@liv.ac.uk'
)
# RUN THIS CODE
'''
python3 setup.py sdist
python3 -m twine upload dist/*

pip uninstall pba
pip install /Users/nickgray/Documents/PhD/code/pba-for-python/dist/pba-0.5.6.tar.gz --user
'''



# import sys
#
# # this is a pointer to the module object instance itself.
# this = sys.modules[__name__]
#
# # we can explicitly make assignments on it
# this.dependancies = None
#
# def initialize():
#     if (this.dependancies is None):
#         # also in local function scope. no scope specifier like global is needed
#         this.dependancies = {}
#         # also the name remains free for local use
#         dependancies = "Locally scoped dependancies variable. Doesn't do anything here."
#     else:
#         msg = "Database is already initialized to {0}."
#         raise RuntimeError(msg.format(this.dependancies))
